module RefacAlphaEq where

import PrettyPrint
import HsLexerPos 
import PosSyntax
import AbstractIO
import Maybe
import TypedIds
import UniqueNames hiding (srcLoc)
import PNT
import HsName
import TiPNT
import List
import RefacUtils hiding (getParams, defines)
import PFE0 (findFile)
import MUtils (( # ))
import RefacLocUtils
import System
import IO

import TransfoUtils

{- Julien Cohen 2010
   This refactoring operation replace a call to a function f into a call to a function g, where f and g are (limited-)alpha-equivalent top-level functions declared in a single module. 
-- WARNING : this operation should be correct in the trivial case. It has not been tested for other cases.
-}

all2 _ [] [] = True
all2 _ [] (_:_) = False
all2 _ (_:_) [] = False
all2 p (e1:r1) (e2:r2) = p e1 e2 && all2 p r1 r2 




limited_alpha_eq:: HsDeclP -> HsDeclP -> Bool
limited_alpha_eq (Dec (HsFunBind _ equations1)) (Dec (HsFunBind _ equations2)) =
    all2 limited_alpha_eq_match equations1 equations2
limited_alpha_eq (Dec (HsPatBind _ pat1 rhs1 localds1)) (Dec (HsPatBind _ pat2 rhs2 localds2)) = error "limited_apha_eq: pat not implemented"
--limited_alpha_eq (Dec (HsPatBind _ _ _ _)) (Dec (HsFunBind _ _))     = error "limited_apha_eq: PatBind Vs FunBind"
--limited_alpha_eq (Dec (HsFunBind _ _))     (Dec (HsPatBind _ _ _ _)) = error "limited_apha_eq: FunBind Vs PatBind"

--limited_alpha_eq (Dec (HsFunBind _ _)) d2 = error (show d2)
--limited_alpha_eq _ _ = error "limited_apha_eq: 1st argument is not a FunBind"

-- Tests whether the declarations are identical up to the renaming of the bound name (only the name of the function).
-- If you want to establish the alpha-equivalence, you can rename all other bound names to make them equals.
-- prerequesites : name1 and name2 are not bound inside the declarations (parameters of the function o in the body or local declarations), free variables must refer to the same definition (or the two declarations have to occur in the same context).
limited_alpha_eq_match (HsMatch _ name1PNT args1 rhs1 localdefs1) (HsMatch _ name2PNT args2 rhs2 localdefs2) =
    let name1 = pNTtoName name1PNT
        name2 = pNTtoName name2PNT

    in  compare_pats args1 args2 && compare_rhs (name1, rhs1) (name2, rhs2) && compare_decls (name1, localdefs1) (name2, localdefs2)

    where
      compare_pats = all2 compare_pat

      compare_pat (Pat(HsPParen p)) p' = compare_pat p p'
      compare_pat p (Pat(HsPParen p')) = compare_pat p p'

      compare_pat (Pat (HsPId i1))              (Pat (HsPId i2))              = compare_ident i1 i2
      compare_pat (Pat (HsPLit _ l1))           (Pat (HsPLit _ l2))           = l1==l2
      compare_pat (Pat (HsPNeg _ l1))           (Pat (HsPNeg _ l2))           = l1==l2
      compare_pat (Pat (HsPSucc _ i1 l1))       (Pat (HsPSucc _ i2 l2))       = l1==l2 && compare_pat_pnt i1 i2
      compare_pat (Pat (HsPInfixApp p1 i1 pp1)) (Pat (HsPInfixApp p2 i2 pp2)) = compare_pat p1 p2 && compare_pat pp1 pp2 && compare_pat_pnt i1 i2
      compare_pat (Pat (HsPApp i1 args1))       (Pat (HsPApp i2 args2))       = compare_pat_pnt i1 i2 && compare_pats args1 args2
      compare_pat (Pat (HsPTuple _ pats1))      (Pat (HsPTuple _ pats2))      = compare_pats pats1 pats2
      compare_pat (Pat (HsPList  _ pats1))      (Pat (HsPList  _ pats2))      = compare_pats pats1 pats2
      compare_pat (Pat (HsPRec _ _))            (Pat (HsPRec _ _))            = error "limited_alpha_eq_match: not implemented on record"
      compare_pat (Pat (HsPAsPat i1 p1))        (Pat (HsPAsPat i2 p2))        = compare_pat_pnt i1 i2 && compare_pat p1 p2
      compare_pat (Pat (HsPWildCard))           (Pat (HsPWildCard))           = True
      compare_pat (Pat (HsPIrrPat p1))          (Pat (HsPIrrPat p2))          = compare_pat p1 p2

      compare_ident (HsVar i1) (HsVar i2) = compare_pat_pnt i1 i2
      compare_ident (HsCon i1) (HsCon i2) = compare_pat_pnt i1 i2
      compare_ident _          _          = False

      compare_pat_pnt (PNT (PN name1 _) _ _) (PNT (PN name2 _) _ _) = name1 == name2
                                                        
      compare_decls (_,[]) (_,[]) = True
      compare_decls _ _ = error "limited_alpha_eq_match: not implemented for functions with local declarations"

      compare_rhs (name1, HsBody e1) (name2, HsBody e2) = compare_expr (name1,e1) (name2,e2)
      compare_rhs (_, HsGuard _) (_, HsGuard _) =  error "limited_alpha_eq_match: not implemented for functions with guards"
      compare_rhs _ _ = False

      compare_pnt (name1, PNT pn1 typ1 _) (name2, PNT pn2 typ2 _) =
          let (n1,n2) = (pNtoName pn1,pNtoName pn2) in
          typ1 == typ2 && (n1 == n2 || (n1 == name1 && n2 == name2)) -- there is the same same or the names are the ones that are being declared.

      compare_expr (name1,Exp e1) (name2, Exp e2) = 
          case (e1,e2) of 

            (HsParen e1, e2) -> compare_expr (name1,e1)  (name2,Exp e2)
            (e1, HsParen e2) -> compare_expr (name1,Exp e1)  (name2,e2)

            (HsId(HsCon (PNT pn1 typ1 _)),HsId(HsCon (PNT pn2 typ2 _))) -> pn1 == pn2 && typ1 == typ2
            (HsId(HsVar pnt1            ),HsId(HsVar pnt2            )) -> compare_pnt (name1, pnt1) (name2,pnt2)
                
            (HsLit _ l1, HsLit _ l2) -> l1 == l2
 
            (HsInfixApp e1 (HsVar op1) e11, HsInfixApp e2 (HsVar op2) e22) -> 
                compare_pnt  (name1, op1) (name2, op2) &&
                compare_expr (name1, e1)  (name2, e2)  &&
                compare_expr (name1, e11) (name2, e22)
            (HsInfixApp e1 (HsCon op1) e11, HsInfixApp e2 (HsCon op2) e22) -> 
                compare_pnt  (name1, op1) (name2, op2) &&
                compare_expr (name1, e1)  (name2, e2)  &&
                compare_expr (name1, e11) (name2, e22)
                                                      
            (HsApp f1 e1, HsApp f2 e2) -> 
                compare_expr (name1, f1)  (name2, f2)  &&
                compare_expr (name1, e1) (name2, e2)

            (HsNegApp _ e1, HsNegApp _ e2) -> 
                compare_expr (name1, e1) (name2, e2)

            (HsLambda pats1 e1, HsLambda pats2 e2) -> 
                compare_pats pats1 pats2  &&
                compare_expr (name1, e1) (name2, e2)
 
            (HsLet ds1 e1, HsLet ds2 e2) -> -- to have a full alpha-equivalence, it would be necessary to take these bindings into account.
                compare_decls (name1, ds1) (name2, ds2)  &&
                compare_expr (name1, e1) (name2, e2)

            (HsIf e11 e12 e13, HsIf e21 e22 e23) -> 
                compare_expr (name1, e11) (name2, e21) &&
                compare_expr (name1, e12) (name2, e22) &&
                compare_expr (name1, e13) (name2, e23)
                                                      
            (HsCase _ _, HsCase _ _ ) ->  error "limited_alpha_eq_match: not implemented for functions with cases."

            (HsDo _ , HsDo _  ) ->  error "limited_alpha_eq_match: not implemented for functions with do."

            (HsTuple t1, HsTuple t2) -> all2 (\ e1 e2 -> compare_expr (name1,e1) (name2,e2)) t1 t2
            (HsList  t1, HsList  t2) -> all2 (\ e1 e2 -> compare_expr (name1,e1) (name2,e2)) t1 t2

            (HsLeftSection  _ _, HsLeftSection  _ _) ->  error "limited_alpha_eq_match: not implemented for functions with LeftSection."
            (HsRightSection _ _, HsRightSection _ _) ->  error "limited_alpha_eq_match: not implemented for functions with RightSection."  
            (HsRecConstr _ _ _, HsRecConstr _ _ _) ->  error "limited_alpha_eq_match: not implemented for functions with records."
            (HsRecUpdate _ _ _, HsRecUpdate _ _ _) ->  error "limited_alpha_eq_match: not implemented for functions with records."
            (HsEnumFrom _, HsEnumFrom _) ->  error "limited_alpha_eq_match: not implemented for functions with enums."
            (HsEnumFromTo _ _ , HsEnumFromTo _ _) ->  error "limited_alpha_eq_match: not implemented for functions with enums."
            (HsEnumFromThen _ _, HsEnumFromThen _ _) ->  error "limited_alpha_eq_match: not implemented for functions with enums."
            (HsEnumFromThenTo _ _ _, HsEnumFromThenTo _ _ _) ->  error "limited_alpha_eq_match: not implemented for functions with enums."
            (HsListComp _ ,HsListComp _) ->  error "limited_alpha_eq_match: not implemented for functions with list comprehensions."
            (HsExpTypeSig _ _ _ _,HsExpTypeSig _ _ _ _) ->  error "limited_alpha_eq_match: not implemented for functions with type-sig-exprs."
            (HsAsPat _ _,HsAsPat _ _) ->  error "limited_alpha_eq_match: not implemented for functions with as-pat-exprs."
            (HsWildCard, HsWildCard) ->  True
            (HsIrrPat _,HsIrrPat _) ->  error "limited_alpha_eq_match: not implemented for functions with IrrPats."
            (_, _) -> False



replace_ident_tok (pnt1,_) (name2,defmod) toks = 
    case (find_toplevel_definition_in_module (pNTtoName pnt1) defmod, find_toplevel_definition_in_module name2 defmod) of
      (Just decl1, Just decl2) ->
          if limited_alpha_eq decl1 decl2 
          then 
              let pnt2 = extract_ident_of_fun_pat_decl decl2 in
              let newpnt = 
                      case (pnt1,pnt2) of
                        (PNT oldname oldtype oldsrc, PNT replacingname replacingtype replacingoccurencesrc) ->
                            PNT replacingname replacingtype oldsrc
              in let (toks',_) = functional_update_toks toks pnt1 newpnt
                 in toks'
          else error ("the replacement is impossible since the two names do not refer to (limited-)alpha-equivalent declarations;\n" ++  (show decl1) ++ "\n" ++ (show decl2))
      (Nothing, _ ) -> error "definition not found"
      (_, Nothing ) -> error "definition not found"

do_replace_ident_pnt (pnt1,mod) (name2,defmod)  toks filename =
         let newtoks = replace_ident_tok (pnt1,mod) (name2,defmod)  toks in 
         do
           writeRefactoredFiles False [((filename,RefacLocUtils.modified),(newtoks,mod))] -- mod is unused in that function

replaceIdent_IDE args
 = do let fileName = ghead "fileName'" args 
          replacingname    = args!!1
          beginRow         = read (args!!2)::Int
          beginCol         = read (args!!3)::Int
          endRow           = read (args!!4)::Int
          endCol           = read (args!!5)::Int

      AbstractIO.putStrLn "do_replace_ident" 
     
      (_, _, mod, tokList)<-parseSourceFile fileName
      let oldVar = locToExp (beginRow, beginCol) (endRow, endCol) tokList mod
          oldVarPNT = case oldVar of
                        Exp (HsId (HsVar pnt)) -> pnt
                        _ -> error "do_replace_ident: unexpected error"
          def_mod_name = case get_definition_module_PNT oldVarPNT of 
                           Just m -> m
                           Nothing -> error "The replaced/replacing identifier must refer to a top-level declaration"

      def_fic <- PFE0.findFile (strToModName def_mod_name)
      (_,_,def_mod_AST,_) <-parseSourceFile (def_fic)
      do_replace_ident_pnt (oldVarPNT,mod) (replacingname,def_mod_AST)  tokList fileName  

replaceIdent_API ident_to_replace_PNT module_name replacingname =
    do
      occurence_fic <- PFE0.findFile (strToModName module_name)
      (_, _, mod, tokList)<-parseSourceFile occurence_fic
      
      let def_mod_name = case get_definition_module_PNT ident_to_replace_PNT of 
                           Just m -> m
                           Nothing -> error "The replaced/replacing identifier must refer to a top-level declaration"


      def_fic <- PFE0.findFile (strToModName def_mod_name)
      (_,_,def_mod_AST,_) <-parseSourceFile (def_fic)
      do_replace_ident_pnt (ident_to_replace_PNT,mod) (replacingname,def_mod_AST)  tokList occurence_fic  

          
            
