module RefacSimplifyCase (
                          simpide,  --interface for IDE
                          simp_op2, --interface for other refactoring operations
                          case_to_eq2 --interface for other refactoring operations
                         ) where

{- Julien Cohen 2010
   The refactoring operation simpide/simp_op2 transforms "case (e1,e2) of (id1,pat2) -> rhs" into "let id1 = e1 in case (e2) of (pat2) -> rhs" .
   The patterns and the matched expression have to be tuples.
   The refactoring applies when there is a same identifier at the same position in all the pattern tuples.
   Input : select the whole case expression.
   Comment 1 : this refactoring is purely structural/syntactical, there is no semantic analysis.
   Comment 2 : This refactoring has not been tested nor proved correct, there is no guaranty it works.
   Comment 3 : Unlike for Haskell Refactorer refactorings, this refactoring source code does not use StrategyLib. We have tried to avoid monads also (just because the author is not easy with it).
-}
   

import PrettyPrint
import HsLexerPos 
import PosSyntax
import AbstractIO
import Maybe
import TypedIds
import UniqueNames hiding (srcLoc)
import PNT
import HsName
import TiPNT
import List
import RefacUtils hiding (getParams, defines)
import PFE0 (findFile)
import MUtils (( # ))
import RefacLocUtils
import System
import IO

import TransfoUtils


{- simpide is called from the IDE. It extracts the expression from the selected region and calls simp_op. -}

simpide args
 = do let fileName = ghead "fileName'" args 
          beginRow         = read (args!!1)::Int
          beginCol         = read (args!!2)::Int
          endRow           = read (args!!3)::Int
          endCol           = read (args!!4)::Int
      AbstractIO.putStrLn "simp_ide" 
     
      (_, _, mod, tokList)<-parseSourceFile fileName
      let oldexp = locToExp (beginRow, beginCol) (endRow, endCol) tokList mod
      simp_op oldexp fileName tokList mod -- here mod should be unused


simp_op2 expr modName = 
    do
      fic <- PFE0.findFile (strToModName modName)
      (_, _, mod, tokList)<-parseSourceFile fic
      simp_op expr fic tokList mod 

{- simp_op takes an expression to simplify and the stram of tokens of the module, 
   computes a new token stream where the given expression has been simplified, and write it into the file. -}
{- The module AST given as argument is unused (see writeRefactoredFiles). -}
simp_op expr filename toklist mod = 
    
         let newtoks = simplify toklist expr in 
         do
           writeRefactoredFiles False [((filename,RefacLocUtils.modified),(newtoks,mod))] -- mod is unused in that function


simplify toks  t@(Exp (HsCase (Exp(HsTuple es)) alts)) =
    case can_simplify t of
      Nothing -> error "No simplification found."
      Just i -> let (lex',alts') = remove_from_alts_LEX_AST i (length es) toks alts
                    (lex'',es') = remove_from_tuple_LEX_AST i (length es) toks es
                    Just id  = sub_pat i (head alts)
                    e = es!!i
                    new_expr = Exp (HsLet [Dec(HsPatBind 
                                               default_srcloc 
                                               (Pat (HsPId (HsVar (nameToPNT id)))) 
                                               (HsBody e)
                                               []
                                              )] 
                                    (Exp (HsCase (Exp(HsTuple es')) alts')))
                    (lex''', _) = functional_update_toks lex'' t new_expr
                in lex''' 
-- c'est peut-être bancal -- vérifier (notamment parce qu'on appelle update sur t alors que t n'existe plus dans lex'' et aussi parce-que on n'a peut-être pas besoin de faire les updates incrémentaux.



    where


      sub_pat n (HsAlt _ (Pat (HsPTuple _ t)) _ _) = 
          case t!!n {- !! starts at 0 -} of
            Pat (HsPId (HsVar i)) -> Just (get_ident_PNT i)
            Pat _ -> error "identifier not found here"

      get_all_subpats n [] = []
      get_all_subpats n (a:as) = (sub_pat n a) : (get_all_subpats n as)

      all_equals [] = True
      all_equals (x:xs) = all (\ y -> x == y) xs

      first_index_all_equal_id max = first_index_aux max 0

      first_index_aux max n = 
          if n == max then Nothing
          else if all_equals (get_all_subpats n alts)
               then Just n
               else first_index_aux max (n+1)


         
      remove_from_alts_LEX_AST n length toks [] = (toks,[])
      remove_from_alts_LEX_AST n length toks (a:as) = 
          let (toks',a') = remove_from_alt_tuple_LEX_AST n length toks a in
          let (toks_res,as') = remove_from_alts_LEX_AST n length toks' as in
          (toks_res, (a':as'))

      remove_from_alt_tuple_LEX_AST n length toks (HsAlt loc1 (Pat (HsPTuple loc2 t)) r d) =
          let (toks',t') = remove_from_tuple_LEX_AST n length toks t in
          (toks', (HsAlt loc1 (Pat (HsPTuple loc2 t')) r d)) -- loc1 and loc2 may not be appropriate

      remove_from_tuple_LEX_AST :: forall t. (Term t, StartEndLoc t, Printable t) => Int -> Int ->  [PosToken] -> [t] -> ( [PosToken] , [t] )
      remove_from_tuple_LEX_AST pos_to_be_removed tuple_length toks ps =
          if pos_to_be_removed <= 0 
          then remove_first_from_tuple_LEX_AST toks ps
          else if pos_to_be_removed >= (tuple_length - 1)
               then remove_last_from_tuple_LEX_AST toks ps
               else remove_middle_from_tuple_LEX_AST pos_to_be_removed toks ps 
          
      remove_first_from_tuple_LEX_AST:: forall t. (Term t, StartEndLoc t, Printable t) =>  [PosToken] -> [t] -> ( [PosToken] , [t] )
      remove_first_from_tuple_LEX_AST toks tuple = 
          let  p = head tuple in 
          let (s,e) = RefacLocUtils.getStartEndLoc toks p in 
          let (a,b,c) = splitToks (s,e) toks in
          let c' = dropWhile isWhite c in
          let c'' = tail c' -- we remove the comma after the ident
          in (a ++ c'', tail tuple)

      remove_last_from_tuple_LEX_AST:: forall t. (Term t, StartEndLoc t, Printable t) =>  [PosToken] -> [t] -> ( [PosToken] , [t] )
      remove_last_from_tuple_LEX_AST toks tuple = 
          let  p = last tuple in 
          let (s,e) = RefacLocUtils.getStartEndLoc toks p in 
          let (a,b,c) = splitToks (s,e) toks in
          let a' = dropWhile isWhite (reverse a) in
          let a'' = tail a' -- we remove the comma before the ident
          in ((reverse a'') ++ c, init tuple)

      remove_middle_from_tuple_LEX_AST:: forall t. (Term t, StartEndLoc t, Printable t) => Int -> [PosToken] -> [t] -> ( [PosToken] , [t] )
      remove_middle_from_tuple_LEX_AST index toks tuple = 
          let p = tuple!!index in
          let (s,e) = RefacLocUtils.getStartEndLoc toks p in 
          let (a,b,c) = splitToks (s,e) toks in
          let c' = dropWhile isWhite c in
          let c'' = tail c' -- we remove the comma after the ident
          in (a ++ c'', remove index tuple)

      remove 0 (x:xs) = xs
      remove n (x:xs) = x : (remove (n-1) xs)

      default_srcloc = SrcLoc "" 0 0 0


      can_simplify (Exp (HsCase (Exp(HsTuple es)) alts)) =
          let n = length es 
          in if all (pat_is_tuple_of_size_n n) alts && n>0
             then  first_index_all_equal_id n 
             else Nothing

          where
            pat_is_tuple_of_size_n n (HsAlt _ (Pat (HsPTuple _ t)) _ _) = (length t == n)

      can_simplify (Exp _) = Nothing



simplify _ t = error ("A case expression on a tuple has to be selected, but the expression selected was :" ++ show t)

       

---------------------------------------------------------------
-- Transfom a case into defining equations
---------------------------------------------------------------


-- transform a match list from a declaration with a case at the top-level into a matching taking the case into account.
match_from_toplevel_case (HsMatch _ _ _ (HsGuard _) _) = error ("This case transformation is not supported when the function has guards.")

match_from_toplevel_case (HsMatch _ funname patlist (HsBody e) ds) =
-- The simplest thing to do with ds is to duplicate it into all the equations.
   case (remove_parents e) of
     Exp (HsCase e' alts) -> 
         case (remove_parents e') of 
           (Exp(HsId (HsVar v))) -> map (fusion v funname patlist ds) alts
           (Exp(HsTuple([Exp(HsId (HsVar v1)),Exp(HsId (HsVar v2))]))) -> map (fusion_double v1 v2 funname patlist ds) alts
           (Exp(HsId (HsCon _))) -> error "This case transformation is not supported when the function contains a case with a matched expression not reduced to a variable which is a parameter of the function."
           (Exp _) -> error "This case transformation is not supported when the function contains a case with a matched expression not reduced to a variable which is a parameter of the function."
     Exp _ -> error "This case transformation is not supported when the function does not have a case-espr at its to-level."

   where fusion v funname patlist ds (HsAlt loc pat body ds') =
                   HsMatch loc funname (replace v pat patlist) body (ds ++ ds') 

         fusion_double v1 v2 funname patlist ds (HsAlt loc pat body ds') =
             let patlist1 = replace v1 (left_pattern pat) patlist in
             let patlist2 = replace v2 (right_pattern pat) patlist1 in
             HsMatch loc funname patlist2 body (ds++ds')

         replace id p ps = map (replace1 id p) ps --only one should change
         
         replace1 id p p'@(Pat (HsPId (HsVar i))) = if same_ident id i then p else p'
         replace1 id p p'@(Pat (HsPId (HsCon _))) = p'                                                    
         replace1 id p p'@(Pat (HsPLit _ _)) = p'
         replace1 id p p'@(Pat (HsPNeg _ _))  = p'
         replace1 id p p'@(Pat (HsPSucc _ _ _)) = p'
         replace1 id p (Pat (HsPInfixApp p1 op p2)) = Pat (HsPInfixApp (replace1 id p p1) op (replace1 id p p2) )
         replace1 id p (Pat (HsPApp i ps)) = Pat (HsPApp i (replace id p ps))
         replace1 id p (Pat (HsPTuple s t)) = Pat (HsPTuple s (replace id p t))
         replace1 id p (Pat (HsPList  s t)) = Pat (HsPList  s (replace id p t))
         replace1 id p (Pat (HsPParen p'))  = Pat (HsPParen (replace1 id p p'))
         replace1 id p (Pat (HsPRec _ _)) = error "This case transformation is not supported with 'rec' patterns."
         replace1 id p (Pat (HsPAsPat i p')) = Pat (HsPAsPat i (replace1 id p p'))
         replace1 id p p'@(Pat (HsPWildCard)) = p'
         replace1 id p (Pat (HsPIrrPat p')) = Pat (HsPIrrPat (replace1 id p p')) -- what is this kind of patterns ?
         

         remove_parents (Exp (HsParen e) ) = remove_parents e
         remove_parents e = e

         left_pattern (Pat (HsPParen p)) = left_pattern p
         left_pattern (Pat (HsPTuple _ [p,_])) = p
         left_pattern _ = error "This case transformation is not supported with such patterns."

         right_pattern (Pat (HsPParen p)) = right_pattern p
         right_pattern (Pat (HsPTuple _ [_,p])) = p
         right_pattern _ = error "This case transformation is not supported with such patterns."

         
case_to_eq d@(Dec (HsFunBind src [m])) filename toks mod = 
    let newdec = Dec (HsFunBind src (match_from_toplevel_case m))
        (toks',_) = functional_update_toks toks d newdec
                    
    in do            writeRefactoredFiles False [((filename,RefacLocUtils.modified),(toks',mod))] -- mod is unused in that function
case_to_eq _ _ _ _ = error "This case transformation is not supported when the function is declared by several equations."

case_to_eq2 expr modName = 
    do
      fic <- PFE0.findFile (strToModName modName)
      (_, _, mod, tokList)<-parseSourceFile fic
      case_to_eq expr fic tokList mod 

{- To test the above operation :

eq  x y = case (x,y) of
            (0,0) -> True
            (0,_) -> False

--->

eq 0 0 = True
eq 0 _ = False


(it should work with patterns of size 1 and 2.)
-}