module TransfoChains where

import PrettyPrint
import PosSyntax
import AbstractIO
import Maybe
import TypedIds
import UniqueNames hiding (srcLoc)
import PNT
import TiPNT
import List
import RefacUtils hiding (getParams) 
import PFE0 (findFile)
import MUtils (( # ))
import qualified RefacLocUtils
import System
import IO
import qualified RefacNewDef
import RefacGenDef
import RefacMoveDef
import RefacMvDefBtwMod
import qualified RefacFunDef -- for subFunctionDef : Fold Definition
import RefacRenaming
--import qualified RefacUnsafeRenaming
import RefacDupDef
import RefacRmDef
import qualified RefacCleanImps
import qualified RefacGenFold
import TransfoUtils -- Part of the transformer layer
import qualified HsLexerPos -- for duplicateInotComment

import qualified TransfoOperations

-- chain
new_def_branches funname prefix modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Just (Dec (HsFunBind _ matchlist)) -> 
            let poslist = map (\(HsMatch _ _ _ rhs _) -> RefacLocUtils.startEndLoc tokList rhs) matchlist
                poslist' = put_numbers 0 poslist
                actionstodo = map (\(((l,c),(l',c')),n) ->  RefacNewDef.introNewDef [fic,prefix ++ show n, show l, show c, show l', show c']) poslist'
            in sequence_ (reverse actionstodo) -- sequence_ executes the monads in its parameter /or transforms a list of monad into the monad for their sequence
             {- Important : there is a trick here. The list has to be reversed. If not, once the first case of the function definition is refactored, all the coordinates of the following cases become invalid. Refactoring from end of the definition to the top allows to avoid to parse the file several times to grep to correct coordiates. -}
        Nothing -> error ("No top-level declaration of " ++ funname ++ " found in " ++ modname ++ ".")
        Just _ -> error ("The definition of " ++ funname ++ " in " ++ modname ++ " must be a function (funbind).")

    where put_numbers _ [] = []
          put_numbers n (e:r) = (e,n) : (put_numbers (n+1) r)



-- chain
generaliseAllRecAppToPatternVariablesInAllLocalDefsIn defname local_funs_prefix new_args_prefix modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let target_decls = find_target_decls_in_funbind_in_module defname local_funs_prefix mod
          l  = map (\d -> ((get_ident_PNT (extract_ident_of_fun_decl d)), (find_app_of defname 1 Nothing d))) (target_decls::[HsDeclP]) -- FIXME : 100 -- contains pairs with a name of  a local decl and a list of applications of interest
          l' = concat (map (\(name,s) -> map (\ _ -> name) s) l) -- l' contains a number of occurences of name which coorespond to the number of applications of defname to generalise in teh local definition of name
          l'' = put_numbers 0 l'
          actions = map (\(name,number) -> TransfoOperations.generalise_fun_app_in_in defname name defname (new_args_prefix ++ show number) modname) l''
          in if l == [] then error "generaliseAllRecAppToPatternVariablesInAllLocalDefsIn : no convenient application found."
             else sequence_ actions 

-- attention, cette façon de faire avec un liste fixée avant de commencer les transformations peut poser problème lorsqu'un generalise va faire disparaître plusieurs instances identiques.

    where put_numbers _ [] = []
          put_numbers n (e:r) = (e,n) : (put_numbers (n+1) r)


-- chain
lift_branch_definitions_to_toplevel funname prefix modname =
    do filename<-PFE0.findFile (strToModName modname)
       (_, _, (mod::HsModuleP), tokList)<-parseSourceFile filename
       let target_decls = find_target_decls_in_funbind_in_module funname prefix mod
           target_positioned_names = map extract_ident_of_fun_decl target_decls 
           coords = map (RefacLocUtils.startEndLoc tokList) target_positioned_names 
           actionstodo = map (\((l,c),(_,_))-> liftToTopLevel [filename, show l, show c]) coords
           in sequence_ (reverse actionstodo)

-- chain
generaliseIdentInAllLocalDefsIn defname local_funs_prefix targetname newname modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let target_decls = find_target_decls_in_funbind_in_module defname local_funs_prefix mod
          -- In each declaration, the generalisation has to be done at most once.
          target_instances = mapMaybe (find_instance_of_ident_in_decl targetname) target_decls
          target_positions = map (RefacLocUtils.startEndLoc tokList) target_instances
          actions = reverse (map (\((l,c),(l',c')) -> RefacGenDef.generaliseDef [fic, newname, show l, show c, show l', show c']) target_positions)
          in sequence_ actions


-- Given a function funname, the parameters (in fact identifiers) of funcame are generalised in the local functions of funname which name begins by prefix.
generaliseAllPatternVariablesInAllLocalDefsIn funname prefix modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_toplevel_definition_in_module funname mod of
        Just (Dec (HsFunBind _ matchlist)) -> 
            let l = map (\(HsMatch _ _ pats _ ds) -> (bound_variables_in_patterns pats,
                                                      filter (is_fun_pat_bind_with_prefix prefix) ds)) 
                        (matchlist::[HsMatchP])

                l' = map (\(pnts,ds) -> (reverse (map (get_ident_PNT) pnts), -- generalise put the new argument at the first position, so we have to reverse here to keep the same order as in the original patterns
                                         map (get_ident_PNT.extract_ident_of_fun_decl) ds)
                         ) l --we keep only names as coordinates will change after each refactoring
                l'' = flat l' -- pairs (list of names, name of a definition)
                in apply_all l''

    where flat [] = []
          flat (([],_):r)      = flat r
          flat ((names,[]):r)  = flat r
          flat ((names, (d:ds)):r) = (names,d) : (flat ( (names,ds) : r))
                  
          apply_all [] = return ()
          apply_all (([],_):r) = apply_all r
          apply_all ( ((name:names),d):r) = do TransfoOperations.generalise_ident_in_in funname d name name  modname ;
                                               apply_all ( (names,d):r)




moveAllDefsToConstructorModulesFollowingPatternMatching motherFunction businessFunctionPrefix sourceModule =
    do
      fic <- PFE0.findFile (strToModName sourceModule)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      case find_business_function_at_toplevel mod businessFunctionPrefix tokList of
        Nothing -> return ()
        Just (name,((l,col),(_,_))) -> 
            case find_toplevel_definition_in_module motherFunction mod of
              Nothing -> error ("moveAllDefsToConstructorModulesFollowingPatternMatching : mother function not found.")
              Just d -> 
                  case find_pattern_constructor_for_case_using_ident d name of
                    Nothing -> error ("moveAllDefsToConstructorModulesFollowingPatternMatching : could not deduce the destination module from the pattern.")
                    Just con -> do 
                      RefacMvDefBtwMod.moveDefBtwMod [fic,(con++"Mod"),show l,show col] -- (con ++ "Mod") is very ad-hoc
                      moveAllDefsToConstructorModulesFollowingPatternMatching motherFunction businessFunctionPrefix sourceModule

    where find_business_function_at_toplevel mod prefix toklist = 
              let decls = get_decl_list mod in     
              case find (is_fun_pat_bind_with_prefix prefix) decls of
                Nothing -> Nothing
                Just d -> let i = extract_ident_of_fun_pat_decl d in  Just  (get_ident_PNT i, RefacLocUtils.startEndLoc toklist i)


          find_pattern_constructor_for_case_using_ident ::  HsDeclP -> String -> Maybe String
          find_pattern_constructor_for_case_using_ident (Dec (HsFunBind _ matches)) i = 
              case find_in_branch matches of 
                Nothing -> Nothing
                Just (HsMatch _ _ ps _ _) -> case constructors_in_patterns ps of
                                               [] -> Nothing
                                               pnt : _ -> Just (get_constructor_ident_PNT pnt)


              where
                find_in_branch [] = Nothing
                find_in_branch (m:ms) = case find_instance_of_ident_in_match i m of
                                          Nothing -> find_in_branch ms
                                          Just _ -> Just m


            
collect_business_functions_and_rename funname businessFunctionPrefix modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::HsModuleP), tokList)<-parseSourceFile fic
      let
          decls         = get_decl_list mod   
          (idents::[PNT])        = extract_identifiers_decls decls 
          target_idents = filter (has_prefix_PNT businessFunctionPrefix) idents 
          (infos::[(String,String)])         = remove_Nothings (nub (map (\ pnt -> (get_ident_PNT pnt, get_definition_module_PNT pnt)) target_idents))
 
      sequence_ (map (\ (i,m) -> TransfoOperations.rename_toplevel (i,m) funname) infos)
                 

    where remove_Nothings [] = []
          remove_Nothings ((i,Nothing):r) = remove_Nothings r
          remove_Nothings ((i,Just m):r)  = (i,m) : (remove_Nothings r)
             