{- Julien Cohen 2010-2011 -}

module TransfoCommands 

{- This file provides an interface to functions defined in TransforOperations (or other).    -}
{- In this interface, all functions have a list of strings as parameter.                     -}
{- The elements of that list have to be passed as separate arguments to interfaced function,
   and, if needed, transformed into integers.                                                -}
{- The list of strings as parameters is needed for interfacing with emacs 
   (at least in the HaRe architecture).                                                      -}


where 

import qualified TransfoOperations
import qualified TransfoChains
import qualified RefacSimplifyCase
--import qualified RefacAlphaEq
import TransfoUtils(CurryMode(..), TypeOfInterest(..))

renameTopLevelIdent args = TransfoOperations.rename_toplevel (args!!0, args!!1) (args!!2)
renameLocalIdentIn args = TransfoOperations.rename_local (args!!0, args!!1) (args!!2)
-- renameTopLevelIdentUnsafe args = TransfoOperations.rename_toplevel_unsafe (args!!0, args!!1) (args!!2)

duplicateTopLevelDefIntoComment args = TransfoOperations.duplicateIntoComment (args!!0) (args!!1)

generativeFoldIdentApp args =
    do 
      a<- TransfoOperations.select_app_of_fun_to_n_args (args!!0) ((read ( args!!1))::Int) (args!!2)
      TransfoOperations.generativeFold a (args!!2)

unfoldIn args = TransfoOperations.unfoldinstance_in (args!!0) (args!!1) (args!!2)

foldToplevelDefinition args = TransfoOperations.fold_toplevel_definition (args!!0) (args!!1)

removeTopLevelDef args = TransfoOperations.removedef  (args!!0) (args!!1)

simplifyCasePatternIn args =
    do
      c <- TransfoOperations.select_case_in (args!!0) (args!!1)
      RefacSimplifyCase.simp_op2 c (args!!1)

removeLocalDefIn args = TransfoOperations.removelocaldef (args!!0) (args!!1) (args!!2)

cleanImportsMod args = TransfoOperations.cleanImports (args!!0)

rmFromExportsMod args = TransfoOperations.rmFromExports (args!!0) (args!!1)

rmCommentBeforeTopLevelDef args = TransfoOperations.rmCommentBefore (args!!0) (args!!1)

caseToEqIn args =
    do
      a <- TransfoOperations.select_toplevel_decl (args!!0) (args!!1)
      RefacSimplifyCase.case_to_eq2 a (args!!1)

newDefEqBodyIn args = TransfoOperations.exhibit_function (args!!0) (args!!1) (args!!2) (args!!3)

generalisePatternVariableInLocalDefIn args = 
    do let funname = args!!0
           constrname = args!!1
           localfunname = args!!2
           modname = args!!3
           (pos::Integer) = read (args!!4)
--           newargname = args!!5
           (curryflag::CurryMode) = if (args!!5) == "Curried" then Curried else UnCurried
       TransfoOperations.generalisePatternVariableInLocalDefIn funname constrname localfunname modname pos {-newargname-} curryflag 

generaliseRecAppToPatternVariableInLocalDefIn args = 
    do let funname = args!!0
           constrname = args!!1
           localfunname = args!!2
           modname = args!!3
           (pos::Integer) = read (args!!4)
           newargname = args!!5
           (curryflag::CurryMode) = if (args!!6) == "Curried" then Curried else UnCurried
       TransfoOperations.generaliseRecAppToPatternVariableInLocalDefIn funname constrname localfunname modname pos newargname curryflag  

generalisePatternVariableInLocalDefIn2 args = 
    do let funname = args!!0
           constrname = args!!1
           localfunname = args!!2
           modname = args!!3
           (pos::Integer) = read (args!!4)
--           newargname = args!!5
           (curryflag::CurryMode) = if (args!!5) == "Curried" then Curried else UnCurried
           (patternposition::Int) =read (args!!6)
       TransfoOperations.generalisePatternVariableInLocalDefIn2 funname constrname localfunname modname pos {- newargname -} curryflag patternposition

generaliseRecAppToPatternVariableInLocalDefIn2 args = 
    do let funname = args!!0
           constrname = args!!1
           localfunname = args!!2
           modname = args!!3
           (pos::Integer) = read (args!!4)
           newargname = args!!5
           (curryflag::CurryMode) = if (args!!6) == "Curried" then Curried else UnCurried
           (patternposition::Int) =read (args!!7)
       TransfoOperations.generaliseRecAppToPatternVariableInLocalDefIn2 funname constrname localfunname modname pos newargname curryflag  patternposition


liftDefToTopLevelIn args = TransfoOperations.make_global_of_local_in (args!!0) (args!!1) (args!!2)
liftOneLevelIn args = TransfoOperations.lift_one_level_in (args!!0) (args!!1) (args!!2)

liftDefToTopLevel args =  TransfoOperations.make_global_of_local (args!!0) (args!!1)

generaliseIdentIn args = TransfoOperations.generaliseident (args!!0, args!!1) (args!!2) (args!!3)
generaliseIdentAppIn args = TransfoOperations.generalise_fun_app_in (args!!0) (args!!1) (read (args!!2)) Nothing (args!!3) (args!!4) 


newDefIdentApp args = TransfoOperations.new_def_fun_app (args!!0) (read (args!!1)) Nothing (args!!2) (args!!3) 
newDefIdentAppIn args = TransfoOperations.new_def_fun_app_in (args!!0) (args!!1) (read (args!!2)) Nothing (args!!3) (args!!4) 
newDefIdentAppToIn args = TransfoOperations.new_def_fun_app_in (args!!0) (args!!1) (read (args!!2)) (Just (args!!3)) (args!!4) (args!!5) 
newDefInfixAppIn args = TransfoOperations.new_def_infix_app_in (args!!0) (args!!1) (args!!2) (args!!3) 
newDefIdentIn args = TransfoOperations.new_def_ident_in (args!!0) (args!!1) (args!!2) (args!!3) 


moveDefBetweenModules args =  TransfoOperations.move_def_between_modules (args!!0) (args!!1) (args!!2)

replaceIdentIn args = TransfoOperations.replaceIdentIn (args!!0) (args!!1) (args!!2) (args!!3) 



-- chains
newDefAllEqBodiesIn args = TransfoChains.new_def_branches (args!!0) (args!!1) (args!!2)
generaliseAllRecAppToPatternVariablesInAllLocalDefsIn args = TransfoChains.generaliseAllRecAppToPatternVariablesInAllLocalDefsIn (args!!0) (args!!1) (args!!2) (args!!3) 
generaliseIdentInAllLocalDefsIn args = TransfoChains.generaliseIdentInAllLocalDefsIn (args!!0) (args!!1) (args!!2) (args!!3) (args!!4) 
generaliseAllPatternVariablesInAllLocalDefsIn args = TransfoChains.generaliseAllPatternVariablesInAllLocalDefsIn (args!!0) (args!!1) (args!!2) 
liftAllLocalDefsToTopLevelIn args = TransfoChains.lift_branch_definitions_to_toplevel (args!!0) (args!!1) (args!!2)
moveAllDefsToConstructorModulesFollowingPatternMatching args =  TransfoChains.moveAllDefsToConstructorModulesFollowingPatternMatching (args!!0) (args!!1) (args!!2)
renameAllIdentsIn args = TransfoChains.collect_business_functions_and_rename (args!!0) (args!!1) (args!!2)
